#pragma once

#include "scripting/ade_api.h"
#include "object/object.h"

namespace scripting {
namespace api {

DECLARE_ADE_OBJ(l_ShipTextures, object_h);

//**********HANDLE: Ship
DECLARE_ADE_OBJ(l_Ship, object_h);
}
}
