#pragma once


void stuff_special_arrival_anchor_name(char *buf, int iff_index, int restrict_to_players, int retail_format);

void stuff_special_arrival_anchor_name(char *buf, int anchor_num, int retail_format);

void generate_weaponry_usage_list_team(int team, int *arr);

void generate_weaponry_usage_list_wing(int wing_num, int *arr);
